@@ -188,6 +188,8 @@ static void as_array_dbl(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr
   // (mostly so that we can support date/time types with various units)
   switch (schema_view->type) {
     case NANOARROW_TYPE_DOUBLE:
+    case NANOARROW_TYPE_FLOAT:
+    case NANOARROW_TYPE_HALF_FLOAT:
     case NANOARROW_TYPE_INT64:
     case NANOARROW_TYPE_INT32:
       break;
@@ -228,7 +230,7 @@ static void as_array_dbl(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr
 
     buffer->size_bytes = len * sizeof(int64_t);
 
-  } else {
+  } else if (schema_view->type == NANOARROW_TYPE_INT32) {
     // double -> int32_t
     struct ArrowBuffer* buffer = ArrowArrayBuffer(array, 1);
     result = ArrowBufferReserve(buffer, len * sizeof(int32_t));
@@ -257,6 +259,23 @@ static void as_array_dbl(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr
     }
 
     buffer->size_bytes = len * sizeof(int32_t);
+  } else {
+    result = ArrowArrayStartAppending(array);
+    if (result != NANOARROW_OK) {
+      Rf_error("ArrowArrayStartAppending() failed");
+    }
+
+    result = ArrowArrayReserve(array, len);
+    if (result != NANOARROW_OK) {
+      Rf_error("ArrowArrayReserve() failed");
+    }
+
+    for (int64_t i = 0; i < len; i++) {
+      result = ArrowArrayAppendDouble(array, x_data[i]);
+      if (result != NANOARROW_OK) {
+        Rf_error("ArrowArrayAppendDouble() failed");
+      }
+    }
   }
 
   // Set the array fields
@@ -301,82 +320,57 @@ static void as_array_dbl(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr
 
 static void as_array_chr(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr,
                          struct ArrowSchemaView* schema_view, struct ArrowError* error) {
-  // Only consider the default create for now
-  if (schema_view->type != NANOARROW_TYPE_STRING) {
-    call_as_nanoarrow_array(x_sexp, array, schema_xptr, "as_nanoarrow_array_from_c");
-    return;
+  switch (schema_view->type) {
+    case NANOARROW_TYPE_BINARY:
+    case NANOARROW_TYPE_STRING:
+    case NANOARROW_TYPE_LARGE_STRING:
+    case NANOARROW_TYPE_LARGE_BINARY:
+    case NANOARROW_TYPE_STRING_VIEW:
+    case NANOARROW_TYPE_BINARY_VIEW:
+      break;
+    default:
+      call_as_nanoarrow_array(x_sexp, array, schema_xptr, "as_nanoarrow_array_from_c");
+      return;
   }
 
   int64_t len = Rf_xlength(x_sexp);
 
-  int result = ArrowArrayInitFromType(array, NANOARROW_TYPE_STRING);
+  int result = ArrowArrayInitFromType(array, schema_view->type);
   if (result != NANOARROW_OK) {
     Rf_error("ArrowArrayInitFromType() failed");
   }
 
-  // Keep these buffers under the umbrella of the array so that we don't have
-  // to worry about cleaning them up if STRING_ELT jumps
-  struct ArrowBuffer* offset_buffer = ArrowArrayBuffer(array, 1);
-  struct ArrowBuffer* data_buffer = ArrowArrayBuffer(array, 2);
-
-  result = ArrowBufferReserve(offset_buffer, (len + 1) * sizeof(int32_t));
+  result = ArrowArrayStartAppending(array);
   if (result != NANOARROW_OK) {
-    Rf_error("ArrowBufferReserve() failed");
+    Rf_error("ArrowArrayStartAppending() failed");
   }
 
-  int64_t null_count = 0;
-  int32_t cumulative_len = 0;
-  ArrowBufferAppendUnsafe(offset_buffer, &cumulative_len, sizeof(int32_t));
-
+  struct ArrowStringView item_view;
   for (int64_t i = 0; i < len; i++) {
     SEXP item = STRING_ELT(x_sexp, i);
+
     if (item == NA_STRING) {
-      null_count++;
+      result = ArrowArrayAppendNull(array, 1);
+      if (result != NANOARROW_OK) {
+        Rf_error("ArrowArrayAppendString() failed");
+      }
     } else {
       const void* vmax = vmaxget();
-      const char* item_utf8 = Rf_translateCharUTF8(item);
-      int64_t item_size = strlen(item_utf8);
-      if ((item_size + cumulative_len) > INT_MAX) {
-        Rf_error("Use na_large_string() to convert character() with total size > 2GB");
-      }
-
-      int result = ArrowBufferAppend(data_buffer, item_utf8, item_size);
+      item_view.data = Rf_translateCharUTF8(item);
+      item_view.size_bytes = strlen(item_view.data);
+      result = ArrowArrayAppendString(array, item_view);
       if (result != NANOARROW_OK) {
-        Rf_error("ArrowBufferAppend() failed");
+        Rf_error("ArrowArrayAppendString() failed");
       }
-      cumulative_len += (int32_t)item_size;
 
       vmaxset(vmax);
     }
-
-    ArrowBufferAppendUnsafe(offset_buffer, &cumulative_len, sizeof(int32_t));
   }
 
-  // Set the array fields
-  array->length = len;
-  array->offset = 0;
-
-  // If there are nulls, pack the validity buffer
-  if (null_count > 0) {
-    struct ArrowBitmap bitmap;
-    ArrowBitmapInit(&bitmap);
-    result = ArrowBitmapReserve(&bitmap, len);
-    if (result != NANOARROW_OK) {
-      Rf_error("ArrowBitmapReserve() failed");
-    }
-
-    for (int64_t i = 0; i < len; i++) {
-      uint8_t is_valid = STRING_ELT(x_sexp, i) != NA_STRING;
-      ArrowBitmapAppendUnsafe(&bitmap, is_valid, 1);
-    }
-
-    ArrowArraySetValidityBitmap(array, &bitmap);
-  }
-
-  array->null_count = null_count;
   result = ArrowArrayFinishBuildingDefault(array, error);
   if (result != NANOARROW_OK) {
-    Rf_error("ArrowArrayFinishBuildingDefault(): %s", error->message);
+    Rf_error("ArrowArrayFinishBuildingDefault() failed with code %d: %s", result,
+             error->message);
   }
 }
 
@@ -428,83 +422,57 @@ static void as_array_data_frame(SEXP x_sexp, struct ArrowArray* array, SEXP sche
 
 static void as_array_list(SEXP x_sexp, struct ArrowArray* array, SEXP schema_xptr,
                           struct ArrowSchemaView* schema_view, struct ArrowError* error) {
-  // We handle list(raw()) in C but fall back to S3 for other types of list output.
-  // Arbitrary nested list support is complicated in C without some concept of a
-  // "builder", which we don't use.
-  if (schema_view->type != NANOARROW_TYPE_BINARY) {
-    call_as_nanoarrow_array(x_sexp, array, schema_xptr, "as_nanoarrow_array_from_c");
-    return;
-  }
-
-  int result = ArrowArrayInitFromType(array, schema_view->type);
-  if (result != NANOARROW_OK) {
-    Rf_error("ArrowArrayInitFromType() failed");
+  switch (schema_view->type) {
+    case NANOARROW_TYPE_BINARY:
+    case NANOARROW_TYPE_LARGE_BINARY:
+    case NANOARROW_TYPE_FIXED_SIZE_BINARY:
+    case NANOARROW_TYPE_BINARY_VIEW:
+      break;
+    default:
+      call_as_nanoarrow_array(x_sexp, array, schema_xptr, "as_nanoarrow_array_from_c");
+      return;
   }
 
   int64_t len = Rf_xlength(x_sexp);
-  struct ArrowBuffer* offset_buffer = ArrowArrayBuffer(array, 1);
-  struct ArrowBuffer* data_buffer = ArrowArrayBuffer(array, 2);
 
-  result = ArrowBufferReserve(offset_buffer, (len + 1) * sizeof(int32_t));
+  // Use schema here to ensure we fixed-size binary byte width works
+  struct ArrowSchema* schema = nanoarrow_schema_from_xptr(schema_xptr);
+  int result = ArrowArrayInitFromSchema(array, schema, error);
   if (result != NANOARROW_OK) {
-    Rf_error("ArrowBufferReserve() failed");
+    Rf_error("ArrowArrayInitFromType() failed: %s", error->message);
   }
 
-  int64_t null_count = 0;
-  int32_t cumulative_len = 0;
-  ArrowBufferAppendUnsafe(offset_buffer, &cumulative_len, sizeof(int32_t));
+  result = ArrowArrayStartAppending(array);
+  if (result != NANOARROW_OK) {
+    Rf_error("ArrowArrayStartAppending() failed");
+  }
 
+  struct ArrowBufferView item_view;
   for (int64_t i = 0; i < len; i++) {
     SEXP item = VECTOR_ELT(x_sexp, i);
+
     if (item == R_NilValue) {
-      ArrowBufferAppendUnsafe(offset_buffer, &cumulative_len, sizeof(int32_t));
-      null_count++;
-      continue;
+      result = ArrowArrayAppendNull(array, 1);
+      if (result != NANOARROW_OK) {
+        Rf_error("ArrowArrayAppendNull() failed");
+      }
+    } else if (TYPEOF(item) == RAWSXP) {
+      item_view.data.data = RAW(item);
+      item_view.size_bytes = Rf_xlength(item);
+      result = ArrowArrayAppendBytes(array, item_view);
+      if (result != NANOARROW_OK) {
+        Rf_error("ArrowArrayAppendBytes() failed");
+      }
+    } else {
+      Rf_error("All list items must be raw() or NULL in conversion to %s",
+               ArrowTypeString(schema_view->type));
     }
-
-    if (Rf_isObject(item) || TYPEOF(item) != RAWSXP) {
-      Rf_error("All list items must be raw() or NULL in conversion to na_binary()");
-    }
-
-    int64_t item_size = Rf_xlength(item);
-    if ((item_size + cumulative_len) > INT_MAX) {
-      Rf_error("Use na_large_binary() to convert list(raw()) with total size > 2GB");
-    }
-
-    result = ArrowBufferAppend(data_buffer, RAW(item), item_size);
-    if (result != NANOARROW_OK) {
-      Rf_error("ArrowBufferAppend() failed");
-    }
-
-    cumulative_len += (int32_t)item_size;
-    ArrowBufferAppendUnsafe(offset_buffer, &cumulative_len, sizeof(int32_t));
   }
 
-  // Set the array fields
-  array->length = len;
-  array->offset = 0;
-
-  // If there are nulls, pack the validity buffer
-  if (null_count > 0) {
-    struct ArrowBitmap bitmap;
-    ArrowBitmapInit(&bitmap);
-    result = ArrowBitmapReserve(&bitmap, len);
-    if (result != NANOARROW_OK) {
-      Rf_error("ArrowBitmapReserve() failed");
-    }
-
-    for (int64_t i = 0; i < len; i++) {
-      uint8_t is_valid = VECTOR_ELT(x_sexp, i) != R_NilValue;
-      ArrowBitmapAppendUnsafe(&bitmap, is_valid, 1);
-    }
-
-    ArrowArraySetValidityBitmap(array, &bitmap);
-  }
-
-  array->null_count = null_count;
   result = ArrowArrayFinishBuildingDefault(array, error);
   if (result != NANOARROW_OK) {
-    Rf_error("ArrowArrayFinishBuildingDefault(): %s", error->message);
+    Rf_error("ArrowArrayFinishBuildingDefault() failed with code %d: %s", result,
+             error->message);
   }
 }
 
