@@ -174,7 +174,7 @@ static void set_converter_list_of(SEXP converter_xptr, struct RConverter* conver
 static int set_converter_children_schema(SEXP converter_xptr, SEXP schema_xptr) {
   struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
-  struct ArrowSchema* schema = schema_from_xptr(schema_xptr);
+  struct ArrowSchema* schema = nanoarrow_schema_from_xptr(schema_xptr);
 
   if (schema->n_children != converter->n_children) {
     ArrowErrorSet(&converter->error,
@@ -201,7 +201,7 @@ static int set_converter_children_schema(SEXP converter_xptr, SEXP schema_xptr)
 static int set_converter_children_array(SEXP converter_xptr, SEXP array_xptr) {
   struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
-  struct ArrowArray* array = array_from_xptr(array_xptr);
+  struct ArrowArray* array = nanoarrow_array_from_xptr(array_xptr);
 
   if (array->n_children != converter->n_children) {
     ArrowErrorSet(&converter->error,
@@ -286,7 +286,7 @@ SEXP nanoarrow_converter_from_ptype(SEXP ptype) {
 int nanoarrow_converter_set_schema(SEXP converter_xptr, SEXP schema_xptr) {
   struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
-  struct ArrowSchema* schema = schema_from_xptr(schema_xptr);
+  struct ArrowSchema* schema = nanoarrow_schema_from_xptr(schema_xptr);
   NANOARROW_RETURN_NOT_OK(
       ArrowSchemaViewInit(&converter->schema_view, schema, &converter->error));
 
@@ -310,7 +310,7 @@ int nanoarrow_converter_set_schema(SEXP converter_xptr, SEXP schema_xptr) {
 int nanoarrow_converter_set_array(SEXP converter_xptr, SEXP array_xptr) {
   struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
-  struct ArrowArray* array = array_from_xptr(array_xptr);
+  struct ArrowArray* array = nanoarrow_array_from_xptr(array_xptr);
   NANOARROW_RETURN_NOT_OK(
       ArrowArrayViewSetArray(&converter->array_view, array, &converter->error));
   SET_VECTOR_ELT(converter_shelter, 2, array_xptr);
@@ -414,11 +414,7 @@ int nanoarrow_converter_finalize(SEXP converter_xptr) {
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
   SEXP current_result = VECTOR_ELT(converter_shelter, 4);
 
-  // Materialize never called (e.g., empty stream)
-  if (current_result == R_NilValue) {
-    NANOARROW_RETURN_NOT_OK(nanoarrow_converter_reserve(converter_xptr, 0));
-    current_result = VECTOR_ELT(converter_shelter, 4);
-  }
+  NANOARROW_RETURN_NOT_OK(nanoarrow_materialize_finalize_result(converter_xptr));
 
   // Check result size. A future implementation could also shrink the length
   // or reallocate a shorter vector.
@@ -437,15 +433,19 @@ int nanoarrow_converter_finalize(SEXP converter_xptr) {
 SEXP nanoarrow_converter_release_result(SEXP converter_xptr) {
   struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
+
   // PROTECT()ing here because we are about to release the object from the
   // shelter of the converter and return it
   SEXP result = PROTECT(VECTOR_ELT(converter_shelter, 4));
   SET_VECTOR_ELT(converter_shelter, 4, R_NilValue);
+
+  // Reset the converter state
   converter->dst.vec_sexp = R_NilValue;
   converter->dst.offset = 0;
   converter->dst.length = 0;
   converter->size = 0;
   converter->capacity = 0;
+
   UNPROTECT(1);
   return result;
 }
