@@ -197,6 +197,46 @@ test_that("as_nanoarrow_array() works for double() -> na_int64()", {
   expect_identical(convert_array(array), as.double(c(1:10, NA_real_)))
 })
 
+test_that("as_nanoarrow_array() works for double() -> na_float()", {
+  # Without nulls
+  array <- as_nanoarrow_array(as.double(1:10), schema = na_float())
+  expect_identical(infer_nanoarrow_schema(array)$format, "f")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  expect_identical(convert_array(array), as.double(1:10))
+
+  # With nulls
+  array <- as_nanoarrow_array(c(1:10, NA_real_), schema = na_float())
+  expect_identical(infer_nanoarrow_schema(array)$format, "f")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 10), FALSE, rep(FALSE, 5)))
+  )
+  expect_identical(convert_array(array), c(1:10, NA_real_))
+})
+
+test_that("as_nanoarrow_array() works for double() -> na_half_float()", {
+  # Without nulls
+  array <- as_nanoarrow_array(as.double(1:10), schema = na_half_float())
+  expect_identical(infer_nanoarrow_schema(array)$format, "e")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  expect_identical(convert_array(array), as.double(1:10))
+
+  # With nulls
+  array <- as_nanoarrow_array(c(1:10, NA_real_), schema = na_half_float())
+  expect_identical(infer_nanoarrow_schema(array)$format, "e")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 10), FALSE, rep(FALSE, 5)))
+  )
+  expect_identical(convert_array(array), c(1:10, NA_real_))
+})
+
 test_that("as_nanoarrow_array() works for integer64() -> na_int32()", {
   skip_if_not_installed("bit64")
 
@@ -295,8 +335,6 @@ test_that("as_nanoarrow_array() works for character() -> na_string()", {
 })
 
 test_that("as_nanoarrow_array() works for character() -> na_large_string()", {
-  skip_if_not_installed("arrow")
-
   # Without nulls
   array <- as_nanoarrow_array(letters, schema = na_large_string())
   expect_identical(infer_nanoarrow_schema(array)$format, "U")
@@ -322,6 +360,37 @@ test_that("as_nanoarrow_array() works for character() -> na_large_string()", {
   )
 })
 
+test_that("as_nanoarrow_array() works for character() -> na_string_view()", {
+  # Without nulls
+  array <- as_nanoarrow_array(letters, schema = na_string_view())
+  expect_identical(infer_nanoarrow_schema(array)$format, "vu")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  # All these strings are shorter than four characters and thus are all inlined
+  expect_identical(length(array$buffers), 3L)
+  expect_identical(as.vector(array$buffers[[3]]), double())
+
+  # With nulls
+  array <- as_nanoarrow_array(c(letters, NA), schema = na_string_view())
+  expect_identical(infer_nanoarrow_schema(array)$format, "vu")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 26), FALSE, rep(FALSE, 5)))
+  )
+  # All these strings are shorter than four characters and thus are all inlined
+  expect_identical(length(array$buffers), 3L)
+  expect_identical(as.vector(array$buffers[[3]]), double())
+
+  # With non-inlinable strings
+  item <- "this string is longer than 12 bytes"
+  array <- as_nanoarrow_array(item, schema = na_string_view())
+  expect_identical(length(array$buffers), 4L)
+  expect_identical(as.raw(array$buffers[[3]]), charToRaw(item))
+  expect_identical(as.vector(array$buffers[[4]]), as.double(nchar(item)))
+})
+
 test_that("as_nanoarrow_array() works for factor() -> na_dictionary()", {
   array <- as_nanoarrow_array(
     factor(letters),
@@ -504,9 +573,36 @@ test_that("as_nanoarrow_array() works for blob::blob() -> na_binary()", {
   )
 })
 
-test_that("as_nanoarrow_array() works for blob::blob() -> na_large_binary()", {
-  skip_if_not_installed("arrow")
+test_that("as_nanoarrow_array() works for blob::blob() -> na_fixed_size_binary()", {
+  # Without nulls
+  array <- as_nanoarrow_array(blob::as_blob(letters), schema = na_fixed_size_binary(1))
+  expect_identical(infer_nanoarrow_schema(array)$format, "w:1")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    as.raw(as_nanoarrow_buffer(paste(letters, collapse = "")))
+  )
 
+  # With nulls
+  array <- as_nanoarrow_array(
+    blob::as_blob(c(letters, NA)),
+    schema = na_fixed_size_binary(1)
+  )
+  expect_identical(infer_nanoarrow_schema(array)$format, "w:1")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 26), FALSE, rep(FALSE, 5)))
+  )
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    c(as.raw(as_nanoarrow_buffer(paste(letters, collapse = ""))), as.raw(0x00))
+  )
+})
+
+test_that("as_nanoarrow_array() works for blob::blob() -> na_large_binary()", {
   # Without nulls
   array <- as_nanoarrow_array(blob::as_blob(letters), schema = na_large_binary())
   expect_identical(infer_nanoarrow_schema(array)$format, "Z")
@@ -535,6 +631,108 @@ test_that("as_nanoarrow_array() works for blob::blob() -> na_large_binary()", {
   )
 })
 
+test_that("as_nanoarrow_array() works for blob::blob() -> na_binary_view()", {
+  # Without nulls
+  array <- as_nanoarrow_array(blob::as_blob(letters), schema = na_binary_view())
+  expect_identical(infer_nanoarrow_schema(array)$format, "vz")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  # All these strings are shorter than four characters and thus are all inlined
+  expect_identical(length(array$buffers), 3L)
+  expect_identical(as.vector(array$buffers[[3]]), double())
+
+  # With nulls
+  array <- as_nanoarrow_array(
+    blob::as_blob(c(letters, NA)),
+    schema = na_binary_view()
+  )
+  expect_identical(infer_nanoarrow_schema(array)$format, "vz")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 26), FALSE, rep(FALSE, 5)))
+  )
+  # All these strings are shorter than four characters and thus are all inlined
+  expect_identical(length(array$buffers), 3L)
+  expect_identical(as.vector(array$buffers[[3]]), double())
+
+  # With non-inlinable strings
+  item <- list(charToRaw("this string is longer than 12 bytes"))
+  array <- as_nanoarrow_array(item, schema = na_binary_view())
+  expect_identical(length(array$buffers), 4L)
+  expect_identical(as.raw(array$buffers[[3]]), item[[1]])
+  expect_identical(as.vector(array$buffers[[4]]), as.double(length(item[[1]])))
+})
+
+test_that("as_nanoarrow_array() works for list(raw()) -> na_binary()", {
+  # Without nulls
+  array <- as_nanoarrow_array(lapply(letters, charToRaw))
+  expect_identical(infer_nanoarrow_schema(array)$format, "z")
+  expect_identical(as.raw(array$buffers[[1]]), raw())
+  expect_identical(array$offset, 0L)
+  expect_identical(array$null_count, 0L)
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    as.raw(as_nanoarrow_buffer(0:26))
+  )
+  expect_identical(
+    as.raw(array$buffers[[3]]),
+    as.raw(as_nanoarrow_buffer(paste(letters, collapse = "")))
+  )
+
+  # With nulls
+  array <- as_nanoarrow_array(c(lapply(letters, charToRaw), list(NULL)))
+  expect_identical(infer_nanoarrow_schema(array)$format, "z")
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    packBits(c(rep(TRUE, 26), FALSE, rep(FALSE, 5)))
+  )
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    as.raw(as_nanoarrow_buffer(c(0:26, 26L)))
+  )
+  expect_identical(
+    as.raw(array$buffers[[3]]),
+    as.raw(as_nanoarrow_buffer(paste(letters, collapse = "")))
+  )
+})
+
+test_that("as_nanoarrow_array() works for list(NULL) -> na_list(na_na())", {
+  array <- as_nanoarrow_array(list(NULL))
+  expect_identical(infer_nanoarrow_schema(array)$format, "+l")
+  expect_identical(array$length, 1L)
+  expect_identical(array$null_count, 1L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    as.raw(as_nanoarrow_array(FALSE)$buffers[[2]])
+  )
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    as.raw(as_nanoarrow_buffer(c(0L, 0L)))
+  )
+  expect_identical(infer_nanoarrow_schema(array$children[[1]])$format, "n")
+  expect_identical(array$children[[1]]$length, 0L)
+})
+
+test_that("as_nanoarrow_array() works for list(integer()) -> na_list(na_int32())", {
+  array <- as_nanoarrow_array(list(1:5, 6:10), schema = na_list(na_int32()))
+  expect_identical(infer_nanoarrow_schema(array)$format, "+l")
+  expect_identical(array$length, 2L)
+  expect_identical(array$null_count, 0L)
+  expect_identical(
+    as.raw(array$buffers[[1]]),
+    as.raw(as_nanoarrow_array(c(TRUE, TRUE))$buffers[[2]])
+  )
+  expect_identical(
+    as.raw(array$buffers[[2]]),
+    as.raw(as_nanoarrow_buffer(c(0L, 5L, 10L)))
+  )
+  expect_identical(infer_nanoarrow_schema(array$children[[1]])$format, "i")
+  expect_identical(array$children[[1]]$length, 10L)
+})
+
 test_that("as_nanoarrow_array() works for unspecified() -> na_na()", {
   skip_if_not_installed("vctrs")
 
